@@ -1,4 +1,4 @@
-/* See LICENSE file for license and copyright information */
+/* SPDX-License-Identifier: Zlib */
 
 #include "session.h"
 
@@ -16,6 +16,7 @@
 #include "utils.h"
 
 #include <glib/gi18n-lib.h>
+#include <pango/pango-font.h>
 #include <stdlib.h>
 
 #ifdef WITH_LIBNOTIFY
@@ -24,8 +25,11 @@
 
 
 static int
-cb_sort_settings(girara_setting_t* lhs, girara_setting_t* rhs)
+cb_sort_settings(const void* data1, const void* data2)
 {
+  const girara_setting_t* lhs = data1;
+  const girara_setting_t* rhs = data2;
+
   return g_strcmp0(girara_setting_get_name(lhs), girara_setting_get_name(rhs));
 }
 
@@ -43,7 +47,7 @@ ensure_gettext_initialized(void)
 static void
 init_template_engine(GiraraTemplate* csstemplate)
 {
-  static const char* variable_names[] = {
+  static const char variable_names[][24] = {
     "session",
     "default-fg",
     "default-bg",
@@ -90,10 +94,16 @@ css_template_fill_font(GiraraTemplate* csstemplate, const char* font)
   girara_template_set_variable_value(csstemplate, "font-family",
       pango_font_description_get_family(descr));
 
-  char* size = g_strdup_printf("%d%s", pango_font_description_get_size(descr) / PANGO_SCALE,
-      pango_font_description_get_size_is_absolute(descr) == FALSE ? "pt" : "");
-  girara_template_set_variable_value(csstemplate, "font-size", size);
-  g_free(size);
+  const int font_size = pango_font_description_get_size(descr);
+  if (!font_size) {
+    girara_debug("no font size given, defaulting to 9");
+    girara_template_set_variable_value(csstemplate, "font-size", "9pt");
+  } else {
+    char* size = g_strdup_printf("%d%s", pango_font_description_get_size(descr) / PANGO_SCALE,
+        pango_font_description_get_size_is_absolute(descr) == FALSE ? "pt" : "");
+    girara_template_set_variable_value(csstemplate, "font-size", size);
+    g_free(size);
+  }
 
   switch (pango_font_description_get_weight(descr)) {
     case PANGO_WEIGHT_THIN:
@@ -168,7 +178,7 @@ fill_template_with_values(girara_session_t* session)
   };
 
   /* parse color values */
-  const char* color_settings[] = {
+  static const char* const color_settings[] = {
     "default-fg",
     "default-bg",
     "inputbar-fg",
@@ -287,7 +297,7 @@ scrolled_window_set_scrollbar_visibility(GtkScrolledWindow* window,
 
 
 girara_session_t*
-girara_session_create()
+girara_session_create(void)
 {
   ensure_gettext_initialized();
 
@@ -310,10 +320,11 @@ girara_session_create()
 
   session_private->elements.statusbar_items = girara_list_new2(
       (girara_free_function_t) girara_statusbar_item_free);
+  g_mutex_init(&session_private->feedkeys_mutex);
 
   /* settings */
   session_private->settings = girara_sorted_list_new2(
-      (girara_compare_function_t) cb_sort_settings,
+      cb_sort_settings,
       (girara_free_function_t) girara_setting_free);
 
   /* CSS style provider */
@@ -351,11 +362,11 @@ girara_session_create()
   girara_config_load_default(session);
 
   /* create widgets */
-  session->gtk.box                      = GTK_BOX(gtk_box_new(GTK_ORIENTATION_VERTICAL, 0));
+  session->gtk.box                = GTK_BOX(gtk_box_new(GTK_ORIENTATION_VERTICAL, 0));
   session_private->gtk.overlay    = gtk_overlay_new();
   session_private->gtk.bottom_box = GTK_BOX(gtk_box_new(GTK_ORIENTATION_VERTICAL, 0));
-  session->gtk.statusbar_entries        = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
-  session->gtk.inputbar_box             = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
+  session->gtk.statusbar_entries  = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
+  session->gtk.inputbar_box       = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
   gtk_box_set_homogeneous(session->gtk.inputbar_box, TRUE);
   session->gtk.view              = gtk_scrolled_window_new(NULL, NULL);
   session->gtk.viewport          = gtk_viewport_new(NULL, NULL);
@@ -373,6 +384,16 @@ girara_session_create()
   return session;
 }
 
+static void
+screen_changed(GtkWidget* widget, GdkScreen* GIRARA_UNUSED(old_screen), gpointer GIRARA_UNUSED(userdata)) {
+  GdkScreen* screen = gtk_widget_get_screen(widget);
+  GdkVisual* visual = gdk_screen_get_rgba_visual(screen);
+  if (!visual) {
+    visual = gdk_screen_get_system_visual(screen);
+  }
+  gtk_widget_set_visual(widget, visual);
+}
+
 bool
 girara_session_init(girara_session_t* session, const char* sessionname)
 {
@@ -380,15 +401,13 @@ girara_session_init(girara_session_t* session, const char* sessionname)
     return false;
   }
 
-  bool smooth_scroll = false;
-  girara_setting_get(session, "smooth-scroll", &smooth_scroll);
-  if (smooth_scroll == true) {
-    gtk_widget_add_events(session->gtk.viewport, GDK_SMOOTH_SCROLL_MASK);
-  }
-
+  /* set session name */
   session->private_data->session_name = g_strdup(
       (sessionname == NULL) ? "girara" : sessionname);
 
+  /* enable smooth scroll events */
+  gtk_widget_add_events(session->gtk.viewport, GDK_SMOOTH_SCROLL_MASK);
+
   /* load CSS style */
   fill_template_with_values(session);
   g_signal_connect(G_OBJECT(session->private_data->csstemplate), "changed",
@@ -407,6 +426,9 @@ girara_session_init(girara_session_t* session, const char* sessionname)
 
   gtk_widget_set_name(session->gtk.window, session->private_data->session_name);
 
+  g_signal_connect(G_OBJECT(session->gtk.window), "screen-changed", G_CALLBACK(screen_changed), NULL);
+  screen_changed(GTK_WIDGET(session->gtk.window), NULL, NULL);
+
   /* apply CSS style */
   css_template_changed(session->private_data->csstemplate, session);
 
@@ -447,12 +469,12 @@ girara_session_init(girara_session_t* session, const char* sessionname)
   char* guioptions = NULL;
   girara_setting_get(session, "guioptions", &guioptions);
 
-  const bool show_hscrollbar = strchr(guioptions, 'h') != NULL;
-  const bool show_vscrollbar = strchr(guioptions, 'v') != NULL;
+  const bool show_hscrollbar = guioptions != NULL && strchr(guioptions, 'h') != NULL;
+  const bool show_vscrollbar = guioptions != NULL && strchr(guioptions, 'v') != NULL;
+  g_free(guioptions);
 
   scrolled_window_set_scrollbar_visibility(
     GTK_SCROLLED_WINDOW(session->gtk.view), show_hscrollbar, show_vscrollbar);
-  g_free(guioptions);
 
   /* viewport */
   gtk_container_add(GTK_CONTAINER(session->gtk.view), session->gtk.viewport);
@@ -473,6 +495,7 @@ girara_session_init(girara_session_t* session, const char* sessionname)
 
   widget_add_class(GTK_WIDGET(session->gtk.inputbar_entry), "bottom_box");
   widget_add_class(session->gtk.notification_text, "bottom_box");
+  widget_add_class(GTK_WIDGET(session->gtk.statusbar_entries), "bottom_box");
 
   session->signals.inputbar_key_pressed = g_signal_connect(
       G_OBJECT(session->gtk.inputbar_entry),
@@ -560,12 +583,10 @@ girara_session_init(girara_session_t* session, const char* sessionname)
 
   char* window_icon = NULL;
   girara_setting_get(session, "window-icon", &window_icon);
-  if (window_icon != NULL) {
-    if (strlen(window_icon) != 0) {
-      girara_setting_set(session, "window-icon", window_icon);
-    }
-    g_free(window_icon);
+  if (window_icon != NULL && strlen(window_icon) != 0) {
+    girara_set_window_icon(session, window_icon);
   }
+  g_free(window_icon);
 
   gtk_widget_grab_focus(GTK_WIDGET(session->gtk.view));
 
@@ -611,6 +632,9 @@ girara_session_private_free(girara_session_private_t* session)
   girara_list_free(session->settings);
   session->settings = NULL;
 
+  /* clean up mutex */
+  g_mutex_clear(&session->feedkeys_mutex);
+
   g_slice_free(girara_session_private_t, session);
 }
 
@@ -681,7 +705,6 @@ girara_libnotify(girara_session_t* session, const char *summary,
   }
 
 #ifdef WITH_LIBNOTIFY
-
   const bool was_initialized = notify_is_initted();
 
   if (was_initialized == false) {
@@ -721,11 +744,8 @@ girara_libnotify(girara_session_t* session, const char *summary,
   if (was_initialized == false) {
     notify_uninit();
   }
-
 #else
-
   girara_notify(session, GIRARA_WARNING, "Girara was compiled without libnotify support.");
-
 #endif
 }
 
@@ -839,11 +859,11 @@ girara_mode_add(girara_session_t* session, const char* name)
   g_return_val_if_fail(name != NULL && name[0] != '\0', FALSE);
 
   girara_mode_t last_index = 0;
-  GIRARA_LIST_FOREACH(session->modes.identifiers, girara_mode_string_t*, iter, mode)
+  GIRARA_LIST_FOREACH_BODY(session->modes.identifiers, girara_mode_string_t*, mode,
     if (mode->index > last_index) {
       last_index = mode->index;
     }
-  GIRARA_LIST_FOREACH_END(session->modes.identifiers, girara_mode_string_t*, iter, mode);
+  );
 
   /* create new mode identifier */
   girara_mode_string_t* mode = g_slice_new(girara_mode_string_t);
@@ -892,9 +912,32 @@ girara_set_window_icon(girara_session_t* session, const char* name)
     return false;
   }
 
-  gtk_window_set_icon_name(GTK_WINDOW(session->gtk.window), name);
+  if (strlen(name) == 0) {
+    girara_warning("Empty icon name.");
+    return false;
+  }
 
-  return true;
+  GtkWindow* window = GTK_WINDOW(session->gtk.window);
+  char* path        = girara_fix_path(name);
+  bool success      = true;
+
+  if (g_file_test(path, G_FILE_TEST_EXISTS) == TRUE) {
+    girara_debug("Loading window icon from file: %s", path);
+
+    GError* error = NULL;
+    success = gtk_window_set_icon_from_file(window, path, &error);
+
+    if (success == false) {
+      girara_debug("Failed to load window icon (file): %s", error->message);
+      g_error_free(error);
+    }
+  } else {
+    girara_debug("Loading window icon with name: %s", name);
+    gtk_window_set_icon_name(window, name);
+  }
+
+  g_free(path);
+  return success;
 }
 
 girara_list_t*
